/*
 * Copyright 1997, 1998 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#ifndef demcomps_h
#define demcomps_h

#include <OverlayUnidraw/ovfile.h>
#include <MapUnidraw/mapcomps.h>
#include <MapUnidraw/mapviews.h>
#include <MapUnidraw/mapscripts.h>

class vtkActor;

class ActorComp : public OverlayComp {
public:
  ActorComp(const char* name, OverlayComp* parent =nil);
  virtual Component* Copy();
  virtual ClassId GetClassId();
  virtual boolean IsA(ClassId);
  char* Name() { return _name; }
  float xloc() { return _xloc; }
  float yloc() { return _yloc; }
  float heading() { return _heading; }
  float xdir() { return _xdir; }
  float ydir() { return _ydir; }
  void xloc(float x) { _xloc = x; }
  void yloc(float y) { _yloc = y; }
  void heading(float h) { _heading = h; }
  void xdir(float xd) { _xdir = xd; }
  void ydir(float yd) { _ydir = yd; }
protected:
  char* _name;
  float _xloc, _yloc, _heading, _xdir, _ydir;
};

class ActorView : public OverlayView {
public:
  ActorView();
  virtual ClassId GetClassId();
  virtual boolean IsA(ClassId);
  virtual void Update();
  virtual Graphic* GetGraphic();
protected:
  vtkActor* _actor;
};

class SurfaceComp : public OverlayFileComp {
public:
  SurfaceComp(const char* path, OverlayComp* parent =nil);
  SurfaceComp(istream&, OverlayComp* parent =nil);
  virtual Component* Copy();
  virtual ClassId GetClassId();
  virtual boolean IsA(ClassId);
protected:
    ParamList* GetParamList();
    void GrowParamList(ParamList*);
    static ParamList* _surface_params;
};

class SurfaceView : public OverlayFileView {
public:
  SurfaceView();
  virtual ClassId GetClassId();
  virtual boolean IsA(ClassId);
  //virtual void Update();
  virtual void Load();
protected:
  //virtual void SetSubject(Component*);
};

class SurfaceScript : public OverlayFileScript {
public:
  SurfaceScript(SurfaceComp* =nil);
  virtual ClassId GetClassId();
  virtual boolean IsA(ClassId);
  static int ReadPathName(istream&, void*, void*, void*, void*);
};

class RoadsComp : public OverlayFileComp {
public:
  RoadsComp(const char* path, OverlayComp* parent =nil);
  RoadsComp(istream&, OverlayComp* parent =nil);
  virtual Component* Copy();
  virtual ClassId GetClassId();
  virtual boolean IsA(ClassId);
protected:
    ParamList* GetParamList();
    void GrowParamList(ParamList*);
    static ParamList* _roads_params;
};

class RoadsView : public OverlayFileView {
public:
  RoadsView();
  virtual ClassId GetClassId();
  virtual boolean IsA(ClassId);
  //virtual void Update();
  virtual void Load();
protected:
  //virtual void SetSubject(Component*);
};

class RoadsScript : public OverlayFileScript {
public:
  RoadsScript(RoadsComp* =nil);
  virtual ClassId GetClassId();
  virtual boolean IsA(ClassId);
  static int ReadPathName(istream&, void*, void*, void*, void*);
};

class DemIdrawComp : public MapIdrawComp {
public:
  DemIdrawComp(const char* pathname =nil, OverlayComp* parent =nil);
  DemIdrawComp(istream&, const char* pathname = nil, OverlayComp* parent =nil);
  virtual Component* Copy();
  virtual ClassId GetClassId();
  virtual boolean IsA(ClassId);
protected:
    ParamList* GetParamList();
    void GrowParamList(ParamList*);
    static ParamList* _dem_idraw_params;
};

class DemIdrawView : public MapIdrawView {
public:
  DemIdrawView(DemIdrawComp* =nil);
  virtual ClassId GetClassId();
  virtual boolean IsA(ClassId);
};

class DemIdrawScript : public MapIdrawScript {
public:
  DemIdrawScript(DemIdrawComp* =nil);
  virtual ClassId GetClassId();
  virtual boolean IsA(ClassId);
  
};

#endif
