/*
 * Copyright (c) 1996 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#ifdef VTK

#include <DemUnidraw/demeditor.h>
#include <DemUnidraw/demclasses.h>
#include <DemUnidraw/demcmds.h>
#include <DemUnidraw/demcomps.h>
#include <DemUnidraw/demfunc.h>
#include <DemUnidraw/demviewer.h>
#include <DemUnidraw/lodactor.h>
#include <DemVtk/demdataset.h>
#include <MapUnidraw/mapcomps.h>
#include <OverlayUnidraw/ovpanner.h>
#include <ComTerp/comterp.h>
#include <IVGlyph/bdvalue.h>
#include <Unidraw/catalog.h>
#include <Unidraw/iterator.h>
#include <Unidraw/unidraw.h>
#include <InterViews/border.h>
#include <InterViews/tray.h>
#include <InterViews/frame.h>
#include <InterViews/glue.h>
#include <Dispatch/rpcstream.h>
#include <Attribute/attribute.h>
#include <Attribute/attrlist.h>
#include <Attribute/attrvalue.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef MAKEDEPEND
#include <vtkDataSetReader.h>
#include <vtkPolyDataMapper.h>
#endif
#if defined(SVR4)
#include <float.h>
#endif

char DemEditor::basename[256] = "";
DataSetLODs DemEditor::dstable[80*80];
FBoxObj DemEditor::dsindex0[80*80];
FBoxObj DemEditor::dsindex1[40*40];
FBoxObj DemEditor::dsindex2[20*20];
FBoxObj DemEditor::dsindex3[10*10];
FBoxObj DemEditor::dsindex4[5*5];
int DemEditor::counter = 0;

implementActionCallback(DemEditor)

DemEditor::DemEditor(OverlayComp* comp, DemKit* ok)
: MapEditor(false, ok)
{
  count = counter++;
  _name = nil;
  _vsocket = nil;
  _height = 5.0;
  Init(comp);
  InitCameraValues();
}

DemEditor::DemEditor(const char* file, DemKit* ok)
  : MapEditor(false, ok)
{
  count = counter++;
  if (file)
    _filename = strdup(file);
  _vsocket = nil;
  _height = 5.0;
    if (file == nil) {
	Init();
    } else {
	Catalog* catalog = unidraw->GetCatalog();
	OverlayComp* comp;

	char* realfile = (char*)file;
	const char* url = strdup(catalog->GetAttribute("url"));
	if (strlen(url) > 0) {
	  if (strstr(url, "file:") == url)
	    realfile = (char*)url + 5;
	}
#if 0
	const char* lod = catalog->GetAttribute("lod");
	if (lod && strcmp(lod,"true")==0) {
	  Init();
	  FILE* fp = fopen(realfile, "r");
	  if (fp) {
	    char pathname[256];
	    char realpathname[256];
	    memset(pathname, 0, 256);
	    memset(realpathname, 0, 256);
	    fgets(pathname, 256, fp);
	    if (strlen(pathname)>0)
	      pathname[strlen(pathname)-1] = '\0';
	    fclose(fp);
	    strcat(realpathname, pathname);
	    if (strlen(realpathname)>0)
	      InitLOD(realpathname, realfile);
	  }
	}
	else {
#endif
	  if (catalog->Retrieve(realfile, (Component*&) comp)) {
	    Init(comp);
	    if (((OverlayComp*)comp)->GetAttributeList()->GetAttr("lod")) {
	      const char* lod = ((OverlayComp*)comp)->GetAttributeList()->GetAttr("lod")->Value()->string_ptr();
	      if (lod && strlen(lod)>0) {
		char lodpath[256];
		lodpath[0] = '\0';
		if (url && strlen(url) > 0) {
		  if (strstr(url, "file:") == url) {
		    strcpy(lodpath, (char*)url + 5);
		    char* lastslashpos = strrchr(lodpath, '/');
		    if (lastslashpos)
		      *(lastslashpos+1) = '\0';
		  }
		}
		strcat(lodpath, lod);

		FILE* fp = fopen(lodpath, "r");
		if (fp) {
		  char pathname[256];
		  char realpathname[256];
		  memset(pathname, 0, 256);
		  memset(realpathname, 0, 256);
		  fgets(pathname, 256, fp);
		  if (strlen(pathname)>0)
		    pathname[strlen(pathname)-1] = '\0';
		  fclose(fp);

		  if (url && strlen(url) > 0) {
		    if (strstr(url, "file:") == url) {
		      strcpy(realpathname, (char*)url + 5);
		      char* lastslashpos = strrchr(realpathname, '/');
		      if (lastslashpos)
			*(lastslashpos+1) = '\0';
		    }
		  }

		  strcat(realpathname, pathname);
		  if (strlen(realpathname)>0)
		    InitLOD(realpathname, lodpath);
		}

	      }
	    }

	  } else {
	    Init();
	    fprintf(stderr, "demview: couldn't open %s\n", realfile);
	  }
	  //}
    }
    InitCameraValues();
}

void DemEditor::InitLOD(char* dirname, char* docname) {
  if (strlen(basename)==0) {
  strcpy(basename, dirname);
#if 0
  // build indices
  char path1[256];
  for (int i = 0; i < 80; i++) {
    sprintf(path1, "%s/%02d/", basename, i);
    for (int j = 0; j < 80; j++) {
      dstable[(i*80)+j].dslod[0] = nil;
      dstable[(i*80)+j].dslod[1] = nil;
      dstable[(i*80)+j].dslod[2] = nil;
      dstable[(i*80)+j].dslod[3] = nil;
      char path2[256];
      sprintf(path2, "%s%02d.s0.vtk", path1, j);
      FILE* fp;
      if (fp = fopen(path2, "r")) {
	char buf[256];
	fgets(buf, 256, fp);
	fgets(buf, 256, fp);
	sscanf(buf, "%f %f %f %f", &dsindex0[(i*80)+j]._left,
	       &dsindex0[(i*80)+j]._right,
	       &dsindex0[(i*80)+j]._bottom,
	       &dsindex0[(i*80)+j]._top);
	fclose(fp);
      }
    }
  }
  for (int i = 0; i < 40; i++) {
    for (int j = 0; j < 40; j++) {
      dsindex1[(i*40)+j] = dsindex0[((i*2)*80)+(j*2)];
      dsindex1[(i*40)+j] = dsindex1[(i*40)+j] + dsindex0[(((i*2)+1)*80)+(j*2)];
      dsindex1[(i*40)+j] = dsindex1[(i*40)+j] + dsindex0[((i*2)*80)+(j*2)+1];
      dsindex1[(i*40)+j] = dsindex1[(i*40)+j] + dsindex0[(((i*2)+1)*80)+(j*2)+1];
    }
  }
  for (int i = 0; i < 20; i++) {
    for (int j = 0; j < 20; j++) {
      dsindex2[(i*20)+j] = dsindex1[((i*2)*40)+(j*2)];
      dsindex2[(i*20)+j] = dsindex2[(i*20)+j] + dsindex1[(((i*2)+1)*40)+(j*2)];
      dsindex2[(i*20)+j] = dsindex2[(i*20)+j] + dsindex1[((i*2)*40)+(j*2)+1];
      dsindex2[(i*20)+j] = dsindex2[(i*20)+j] + dsindex1[(((i*2)+1)*40)+(j*2)+1];
    }
  }
  for (int i = 0; i < 10; i++) {
    for (int j = 0; j < 10; j++) {
      dsindex3[(i*10)+j] = dsindex2[((i*2)*20)+(j*2)];
      dsindex3[(i*10)+j] = dsindex3[(i*10)+j] + dsindex2[(((i*2)+1)*20)+(j*2)];
      dsindex3[(i*10)+j] = dsindex3[(i*10)+j] + dsindex2[((i*2)*20)+(j*2)+1];
      dsindex3[(i*10)+j] = dsindex3[(i*10)+j] + dsindex2[(((i*2)+1)*20)+(j*2)+1];
    }
  }
  for (int i = 0; i < 5; i++) {
    for (int j = 0; j < 5; j++) {
      dsindex4[(i*5)+j] = dsindex3[((i*2)*10)+(j*2)];
      dsindex4[(i*5)+j] = dsindex4[(i*5)+j] + dsindex3[(((i*2)+1)*10)+(j*2)];
      dsindex4[(i*5)+j] = dsindex4[(i*5)+j] + dsindex3[((i*2)*10)+(j*2)+1];
      dsindex4[(i*5)+j] = dsindex4[(i*5)+j] + dsindex3[(((i*2)+1)*10)+(j*2)+1];
    }
  }

  // print indices
  FILE* ifp = fopen("dsindex", "w");
  for (int i = 0; i < 80; i++) {
    for (int j = 0; j < 80; j++) {
      fprintf(ifp, "%d %d %f %f %f %f\n", i, j,
	      dsindex0[(i*80)+j]._left,
	      dsindex0[(i*80)+j]._right,
	      dsindex0[(i*80)+j]._bottom,
	      dsindex0[(i*80)+j]._top);
    }
  }
  for (int i = 0; i < 40; i++) {
    for (int j = 0; j < 40; j++) {
      fprintf(ifp, "%d %d %f %f %f %f\n", i, j,
	      dsindex1[(i*40)+j]._left,
	      dsindex1[(i*40)+j]._right,
	      dsindex1[(i*40)+j]._bottom,
	      dsindex1[(i*40)+j]._top);
    }
  }
  for (int i = 0; i < 20; i++) {
    for (int j = 0; j < 20; j++) {
      fprintf(ifp, "%d %d %f %f %f %f\n", i, j,
	      dsindex2[(i*20)+j]._left,
	      dsindex2[(i*20)+j]._right,
	      dsindex2[(i*20)+j]._bottom,
	      dsindex2[(i*20)+j]._top);
    }
  }
  for (int i = 0; i < 10; i++) {
    for (int j = 0; j < 10; j++) {
      fprintf(ifp, "%d %d %f %f %f %f\n", i, j,
	      dsindex3[(i*10)+j]._left,
	      dsindex3[(i*10)+j]._right,
	      dsindex3[(i*10)+j]._bottom,
	      dsindex3[(i*10)+j]._top);
    }
  }
  for (int i = 0; i < 5; i++) {
    for (int j = 0; j < 5; j++) {
      fprintf(ifp, "%d %d %f %f %f %f\n", i, j,
	      dsindex4[(i*5)+j]._left,
	      dsindex4[(i*5)+j]._right,
	      dsindex4[(i*5)+j]._bottom,
	      dsindex4[(i*5)+j]._top);
    }
  }
  fclose(ifp);
#endif
  // load indices
  FILE* ifp = fopen(docname, "r");
  char buf[256];
  fgets(buf, 256, ifp);
  int ii, ij;
  for (int i = 0; i < 80; i++) {
    for (int j = 0; j < 80; j++) {
      fscanf(ifp, "%d %d %f %f %f %f\n", &ii, &ij,
	      &dsindex0[(i*80)+j]._left,
	      &dsindex0[(i*80)+j]._right,
	      &dsindex0[(i*80)+j]._bottom,
	      &dsindex0[(i*80)+j]._top);
    }
  }
  for (int i = 0; i < 40; i++) {
    for (int j = 0; j < 40; j++) {
      fscanf(ifp, "%d %d %f %f %f %f\n", &ii, &ij,
	      &dsindex1[(i*40)+j]._left,
	      &dsindex1[(i*40)+j]._right,
	      &dsindex1[(i*40)+j]._bottom,
	      &dsindex1[(i*40)+j]._top);
    }
  }
  for (int i = 0; i < 20; i++) {
    for (int j = 0; j < 20; j++) {
      fscanf(ifp, "%d %d %f %f %f %f\n", &ii, &ij,
	      &dsindex2[(i*20)+j]._left,
	      &dsindex2[(i*20)+j]._right,
	      &dsindex2[(i*20)+j]._bottom,
	      &dsindex2[(i*20)+j]._top);
    }
  }
  for (int i = 0; i < 10; i++) {
    for (int j = 0; j < 10; j++) {
      fscanf(ifp, "%d %d %f %f %f %f\n", &ii, &ij,
	      &dsindex3[(i*10)+j]._left,
	      &dsindex3[(i*10)+j]._right,
	      &dsindex3[(i*10)+j]._bottom,
	      &dsindex3[(i*10)+j]._top);
    }
  }
  for (int i = 0; i < 5; i++) {
    for (int j = 0; j < 5; j++) {
      fscanf(ifp, "%d %d %f %f %f %f\n", &ii, &ij,
	      &dsindex4[(i*5)+j]._left,
	      &dsindex4[(i*5)+j]._right,
	      &dsindex4[(i*5)+j]._bottom,
	      &dsindex4[(i*5)+j]._top);
    }
  }
  fclose(ifp);
  }  
}

void DemEditor::LoadActorsByFootprint(float* cx, float* cy) {
  FFillPolygonObj fp(cx, cy, 3);
  for (int i4 = 0; i4 < 5; i4++) {
    for (int j4 = 0; j4 < 5; j4++) {
      if (fp.Intersects(dsindex4[(i4*5)+j4])) {
	for (int i3 = i4*2; i3 <= (i4*2)+1; i3++) {
	  for (int j3 = j4*2; j3 <= (j4*2)+1; j3++) {
	    if (fp.Intersects(dsindex3[(i3*10)+j3])) {
	      for (int i2 = i3*2; i2 <= (i3*2)+1; i2++) {
		for (int j2 = j3*2; j2 <= (j3*2)+1; j2++) {
		  if (fp.Intersects(dsindex2[(i2*20)+j2])) {
		    for (int i1 = i2*2; i1 <= (i2*2)+1; i1++) {
		      for (int j1 = j2*2; j1 <= (j2*2)+1; j1++) {
			if (fp.Intersects(dsindex1[(i1*40)+j1])) {
			  for (int i0 = i1*2; i0 <= (i1*2)+1; i0++) {
			    for (int j0 = j1*2; j0 <= (j1*2)+1; j0++) {
			      if (fp.Intersects(dsindex0[(i0*80)+j0])) {
				// make vtk actor for this one
				LODActor* actor = LODActor::New();
				actor->SetXY(i0, j0);
				((DemViewer*)GetViewer())->AddActor(actor);
			      }
			    }
			  }
			}
		      }
		    }
		  }
		}
	      }
	    }
	  }
	}
      }
    }
  }
}

vtkDataSet* DemEditor::GetDataSet(int i, int j, int lod) {
  if (!dstable[i*80+j].dslod[lod]) {
    vtkDataSetReader* reader = vtkDataSetReader::New();
    char dspath[256];
    sprintf(dspath, "%s/%02d/%02d.s%d.vtk", basename, i, j, lod);
    reader->SetFileName(dspath);
    reader->Update();
    vtkPolyData* data = (vtkPolyData*)reader->GetOutput();
    if (!data) {
      return nil;
    }
    DemDataSet* demds = DemDataSet::New();
    demds->SetPolyData(data);
    dstable[i*80+j].dslod[lod] = demds;
  }
  return dstable[i*80+j].dslod[lod];
}

void DemEditor::GetDataSetXY(float px, float py, int*& x, int*& y, int& n) {
  n = 0;
  x = new int[4];
  y = new int[4];
  FPointObj fp(px, py);
  for (int i4 = 0; i4 < 5; i4++) {
    for (int j4 = 0; j4 < 5; j4++) {

      if (dsindex4[(i4*5)+j4].Contains(fp)) {

	for (int i3 = i4*2; i3 <= (i4*2)+1; i3++) {
	  for (int j3 = j4*2; j3 <= (j4*2)+1; j3++) {

	    if (dsindex3[(i3*10)+j3].Contains(fp)) {

	      for (int i2 = i3*2; i2 <= (i3*2)+1; i2++) {
		for (int j2 = j3*2; j2 <= (j3*2)+1; j2++) {

		  if (dsindex2[(i2*20)+j2].Contains(fp)) {

		    for (int i1 = i2*2; i1 <= (i2*2)+1; i1++) {
		      for (int j1 = j2*2; j1 <= (j2*2)+1; j1++) {

			if (dsindex1[(i1*40)+j1].Contains(fp)) {

			  for (int i0 = i1*2; i0 <= (i1*2)+1; i0++) {
			    for (int j0 = j1*2; j0 <= (j1*2)+1; j0++) {

			      if (dsindex0[i0*80+j0].Contains(fp)) {
				x[n] = i0;
				y[n] = j0;
				n++;
			      }
			    }
			  }
			}
		      }
		    }
		  }
		}
	      }
	    }
	  }
	}
      }
    }
  }

}

FBoxObj DemEditor::GetBox(int i, int j) {
  return dsindex0[i*80+j];
}

void DemEditor::InitCameraValues() {
  _CameraX = new BoundedValue(FLT_MIN, FLT_MAX, 10.0, 50.0, 0.0, "%8.2f");
  _CameraY = new BoundedValue(FLT_MIN, FLT_MAX, 10.0, 50.0, 0.0, "%8.2f");
  _CameraZ = new BoundedValue(FLT_MIN, FLT_MAX, 10.0, 50.0, 0.0, "%8.2f");
  _FocalPtX = new BoundedValue(FLT_MIN, FLT_MAX, 10.0, 50.0, 0.0, "%8.2f");
  _FocalPtY = new BoundedValue(FLT_MIN, FLT_MAX, 10.0, 50.0, 0.0, "%8.2f");
  _FocalPtZ = new BoundedValue(FLT_MIN, FLT_MAX, 10.0, 50.0, 0.0, "%8.2f");
  _ClipFront = new BoundedValue(FLT_MIN, FLT_MAX, 10.0, 50.0, 0.0, "%8.2f");
  _ClipBack = new BoundedValue(FLT_MIN, FLT_MAX, 10.0, 50.0, 0.0, "%8.2f");
  _ViewUpX = new BoundedValue(FLT_MIN, FLT_MAX, 10.0, 50.0, 0.0, "%1.6f");
  _ViewUpY = new BoundedValue(FLT_MIN, FLT_MAX, 10.0, 50.0, 0.0, "%1.6f");
  _ViewUpZ = new BoundedValue(FLT_MIN, FLT_MAX, 10.0, 50.0, 0.0, "%1.6f");
  _ViewAngle = new BoundedValue(FLT_MIN, FLT_MAX, 10.0, 50.0, 0.0, "%4.2f");
  _Distance = new BoundedValue(FLT_MIN, FLT_MAX, 10.0, 50.0, 0.0, "%8.2f");
}

DemEditor::DemEditor(boolean initflag, DemKit* ok)
: MapEditor(initflag, ok)
{
  count = counter++;
  _name = nil;
  _vsocket = nil;
  _height = 5.0;
}

DemEditor::~DemEditor() {}

void DemEditor::AddCommands(ComTerp* comterp) {

    comterp->add_command("vhclview", new VehicleFunc(comterp, this));
    comterp->add_command("moveview", new MoveToFunc(comterp, this));
    MapEditor::AddCommands(comterp);
}

void DemEditor::Update() {
    DemViewer* v;

    for (int i = 0; (v = (DemViewer*)GetViewer(i)) != nil; ++i) {
        v->Update();
    }
}

StateVar* DemEditor::GetState (const char* name) {
  return nil;
}

void DemEditor::InitCommands() {
  Catalog* catalog = unidraw->GetCatalog();
  const char* nofp = catalog->GetAttribute("nofp");
  if (strcmp(nofp, "true")==0)
    LODActor::UseFootprint(0);
  else
    LODActor::UseFootprint(1);

  const char* dem = catalog->GetAttribute("dem");
  if (dem && strcmp(dem,"")!=0) {
    DemLoadDemCmd* ldcmd = new DemLoadDemCmd(dem);
    ldcmd->SetEditor(this);
    ldcmd->Execute();
  }
  const char* mapf = catalog->GetAttribute("mapf");
  if (mapf && strcmp(mapf,"")!=0) {
    DemLoadFCCmd* lfcmd = new DemLoadFCCmd(mapf);
    lfcmd->SetEditor(this);
    lfcmd->Execute();
  }
    Attribute* lod = ((OverlayComp*)GetComponent())->GetAttributeList()->GetAttr("lod");
  if (lod) {
    // try to have an initial actor
    LODActor* actor = LODActor::New();
    actor->SetXY(0, 0);
    DemDataSet* demds = (DemDataSet*)GetDataSet(0, 0, 3);
    vtkPolyDataMapper* mapper = vtkPolyDataMapper::New();
    WaterLandLookupTable* table = WaterLandLookupTable::New();
    mapper->SetLookupTable(table);
    mapper->SetInput(demds);
    actor->SetMapper(mapper);
    actor->GetProperty()->SetAmbient(0.15);
    actor->GetProperty()->SetDiffuse(0.4);
    actor->GetProperty()->SetAmbientColor(255.0/255.0, 218.0/255.0, 185.0/255.0);
    ((DemViewer*)GetViewer())->AddActor(actor);
    float bounds[6];
    demds->GetBounds(bounds);
    DemCameraCmd* ccmd = new DemCameraCmd(nil);
    ccmd->SetEditor(this);
    ccmd->set_camera(bounds[0], bounds[2], bounds[5], bounds[1], bounds[3],
		     bounds[5]);
  }
  ComEditor::InitCommands();

  DemIdrawView* vview = (DemIdrawView*)GetViewer()->GetGraphicView();
  Iterator it;
  for (vview->First(it); !vview->Done(it); vview->Next(it)) {
    GraphicView* view = vview->GetView(it);
    if (view->IsA(SURFACE_VIEW))
      ((SurfaceView*)view)->Load();
    else if (view->IsA(ROADS_VIEW))
      ((RoadsView*)view)->Load();
  }

}

void DemEditor::Init(OverlayComp* comp, const char* name) {
    if (!comp) comp = new OverlayIdrawComp;
    _overlay_kit->Init(comp, name);
}

Interactor* DemEditor::Interior() {
    HBorder* hborder = new HBorder;
    VBorder* vborder = new VBorder;
    int gap = round(.1*ivcm);

    _tray->HBox(_tray, _viewer, _tray);

    _tray->VBox(_tray, _viewer, _tray);

    return _tray;
}

void DemEditor::CameraUp() {
    DemCameraUpCmd* upcmd = new DemCameraUpCmd();
    upcmd->SetEditor(this);
    upcmd->Execute();
    delete upcmd;
}

void DemEditor::CameraDown() {
    DemCameraDownCmd* dncmd = new DemCameraDownCmd();
    dncmd->SetEditor(this);
    dncmd->Execute();
    delete dncmd;
}

void DemEditor::CameraYawLeft() {
    DemCameraYawLeftCmd* ylcmd = new DemCameraYawLeftCmd();
    ylcmd->SetEditor(this);
    ylcmd->Execute();
    delete ylcmd;
}

void DemEditor::CameraYawRight() {
    DemCameraYawRightCmd* yrcmd = new DemCameraYawRightCmd();
    yrcmd->SetEditor(this);
    yrcmd->Execute();
    delete yrcmd;
}

#endif
